/*******************************************************************************
 *
 * @author Karolina Krolik <<ahref='mailto:k.krolik@samsung.com'>k.krolik@samsung.com</a>>,
 *         Tadeusz Wlodarkiewicz <<ahref='mailto:t.wlodarkiew@samsung.com'>t.wlodarkiew@samsung.com</a>>
 *
 *
 * ********************************************************************************************************
 *
 * Copyright (c) 2012 Samsung Electronics All Rights Reserved.
 *
 ******************************************************************************/
// name of the players
playerColors = {
    1 : 'blue',
    2 : 'red'
};

/**
 * Implements Box2D library for simulating environment and spheres physics.
 * Handles game logic: turns, sphere positions, score and multitouch. Orders
 * webgl module what and when to draw.
 */
game = function($) {
    // debug mode turned on/off
    var debug;

    // main canvas
    var canvas;

    // number of moves player can make
    var MAX_MOVES = 3;

    // scale for debug mode
    var SCALE = 30;

    // velocity below which object is treated as not moving
    var moveTreshold = 0.05;

    // arrays for storing b2Bodys
    var spheresP1 = [], spheresP2 = [], mainSphere = [];

    // round of moves, incremented after second player makes a move
    var round = 0;

    // types of spheres (for WebGl module)
    var sphereType = {
        MAIN : 0,
        PLAYER1_ACTIVE : 1,
        PLAYER1_IDLE : 2,
        PLAYER2_ACTIVE : 3,
        PLAYER2_IDLE : 4
    };

    // players
    var player = {
        PLAYER1 : 1,
        PLAYER2 : 2
    };

    var gameModes = {
        VIEW : 1,
        PLAY : 2
    };

    var playerColors = {
        1 : 'blue',
        2 : 'red'
    };

    var topBarHeight = 15;

    return {

        /**
         * Initialize game logic and box2d object
         *
         * @param deb
         *            Debug mode. If true debug mode is enabled
         */
        initialize : function(deb) {

            debug = deb;
            canvas = document.getElementById('canvas');
            canvas.width = window.innerWidth;
            canvas.height = window.innerHeight - topBarHeight;

            // disable bottom bar scrolling
            bottomBar = document.getElementById('bottomBar');
            bottomBar.addEventListener("touchstart", function(e) {
                e.preventDefault();
            }, false);

            bottomBar.addEventListener("touchend", function(e) {
                e.preventDefault();
            }, false);

            if (!debug) {
                webgl.webGLStart(canvas);
            }

            // canvas width and height in meters
            var height = canvas.height / SCALE;
            var width = canvas.width / SCALE;

            var b2Vec2 = Box2D.Common.Math.b2Vec2, b2AABB = Box2D.Collision.b2AABB, b2BodyDef = Box2D.Dynamics.b2BodyDef, b2Body = Box2D.Dynamics.b2Body, b2FixtureDef = Box2D.Dynamics.b2FixtureDef, b2World = Box2D.Dynamics.b2World, b2PolygonShape = Box2D.Collision.Shapes.b2PolygonShape, b2CircleShape = Box2D.Collision.Shapes.b2CircleShape, b2DebugDraw = Box2D.Dynamics.b2DebugDraw;

            var world = new b2World(
            // gravity
            new b2Vec2(0, 0),

            // allow sleep
            true);

            var fixDef = new b2FixtureDef;
            fixDef.density = 1.0;
            fixDef.friction = 0.5;
            fixDef.restitution = 0.2;

            var bodyDef = new b2BodyDef;

            // box2d object sizes
            var borderSize = height / 50;
            var playerSphereRadius = width / 15;
            var mainSphereRadius = width / 10;
            var positionP1 = width / 6;
            var positionP2 = positionP1 + width / 2;
            var playerShift = width / 6;

            // create ground
            bodyDef.type = b2Body.b2_staticBody;
            fixDef.shape = new b2PolygonShape;
            fixDef.shape.SetAsBox(width, borderSize);
            bodyDef.position.Set(0, 0);
            world.CreateBody(bodyDef).CreateFixture(fixDef);
            bodyDef.position.Set(0, height);
            world.CreateBody(bodyDef).CreateFixture(fixDef);
            fixDef.shape.SetAsBox(borderSize, height);
            bodyDef.position.Set(0, 0);
            world.CreateBody(bodyDef).CreateFixture(fixDef);
            bodyDef.position.Set(width, 0);
            world.CreateBody(bodyDef).CreateFixture(fixDef);

            // create the boules and the jack
            bodyDef.type = b2Body.b2_dynamicBody;
            bodyDef.linearDamping = 1.0;
            bodyDef.angularDamping = 1.0;

            // create main ball
            fixDef.shape = new b2CircleShape(mainSphereRadius);
            bodyDef.position.x = width / 2;
            bodyDef.position.y = height - mainSphereRadius * 2;
            var sphereBody = world.CreateBody(bodyDef);
            sphereBody.SetUserData(sphereType.MAIN);
            mainSphere.push(sphereBody);
            sphereBody.CreateFixture(fixDef);
            if (!debug) {
                webgl.initSphere(mainSphereRadius / width, sphereType.MAIN);
            }

            // create player 1 spheres
            for ( var i = 0; i < 2; ++i) {
                fixDef.shape = new b2CircleShape(playerSphereRadius);
                bodyDef.position.x = positionP1 + i * playerShift;
                bodyDef.position.y = playerSphereRadius + borderSize;
                sphereBody = world.CreateBody(bodyDef);
                sphereBody.SetUserData(sphereType.PLAYER1_ACTIVE);
                spheresP1.push(sphereBody);
                sphereBody.CreateFixture(fixDef);
                if (!debug) {
                    webgl.initSphere(playerSphereRadius / width, sphereType.PLAYER1_ACTIVE);
                    webgl.initSphere(playerSphereRadius / width, sphereType.PLAYER1_IDLE);
                }
            }

            // create player 2 spheres
            for (i = 0; i < 2; ++i) {
                fixDef.shape = new b2CircleShape(playerSphereRadius);
                bodyDef.position.x = positionP2 + i * playerShift;
                bodyDef.position.y = playerSphereRadius + borderSize;
                sphereBody = world.CreateBody(bodyDef);
                sphereBody.SetUserData(sphereType.PLAYER2_IDLE);
                spheresP2.push(sphereBody);
                sphereBody.CreateFixture(fixDef);
                if (!debug) {
                    webgl.initSphere(playerSphereRadius / width, sphereType.PLAYER2_ACTIVE);
                    webgl.initSphere(playerSphereRadius / width, sphereType.PLAYER2_IDLE);
                }
            }

            // setup debug draw
            if (debug) {
                var debugDraw = new b2DebugDraw();
                debugDraw.SetSprite(document.getElementById("canvas").getContext("2d"));
                debugDraw.SetDrawScale(SCALE);
                debugDraw.SetFillAlpha(0.5);
                debugDraw.SetLineThickness(1.0);
                debugDraw.SetFlags(b2DebugDraw.e_shapeBit | b2DebugDraw.e_jointBit);
                world.SetDebugDraw(debugDraw);

            }

            // current touch position
            var touchX, touchY;
            // position of touch on current move start (the point user touched
            // first)
            var touchStartX, touchStartY;
            // currently selected body object
            var selectedBody;
            // temporary variable indicating that user has started aiming, used
            // for getting the selected body
            var newMouseMove = false;

            var singleTap = false;

            // additional distance of the viewer from the board
            var distance = 0;
            // variable that helps calculating distance change
            var deltaDistance = 0;
            // turning board along the x axis, based on the y-axis finger slide
            var tiltAngle = 0;
            var yPosTilt = 0;
            // twisting board along the z axis based on the x-axis finger slide
            var twistAngle = 0;
            var xPosTwist = 0;

            // defines how much finger can be moved to classify mova as a tap
            singleTapSensitivity = 0.5;

            // temporary variable indicating that user has stopped aiming
            var mouseMoveEnd = false;

            // indicates whether the current move is the last one
            var lastMove = false;

            // player currently playing
            var currentPlayer = player.PLAYER1;

            // set bottom bar initial value
            $("#bottomBar").text(
                    "Rotate/zoom disabled. Player " + playerColors[currentPlayer] + " move");

            // set game mode initial value
            var gameMode = gameModes.PLAY;

            var body;
            var handler;

            $("#game").on("pageshow", function(event) {

                handler = window.setInterval(update, 1000 / 120);

            });

            canvas.addEventListener("touchmove", handleTouchMove, true);

            canvas.addEventListener("touchstart", function(e) {
                if (e.touches.length === 1) {
                    singleTap = true;
                } else {
                    singleTap = false;
                }

                if (e.touches.length === 1) {
                    newMouseMove = true;
                    handleTouchMove(e);
                    touchStartX = e.changedTouches[0].pageX / SCALE;
                    touchStartY = e.changedTouches[0].pageY / SCALE;

                }
                if (gameMode === gameModes.VIEW) {
                    // handling screen zoom and rotation when using 2 or 3
                    // fingers
                    if (e.touches.length === 2) {
                        deltaDistance = Math.sqrt(Math.pow(e.touches[0].clientX
                                - e.touches[1].clientX, 2)
                                + Math.pow(e.touches[0].clientY - e.touches[1].clientY, 2));
                    } else if (e.touches.length === 3) {
                        yPosTilt = e.touches[2].clientY;
                        xPosTwist = e.touches[2].clientX;
                    }

                }
            }, false);

            canvas
                    .addEventListener(
                            "touchend",
                            function(e) {
                                if (gameMode === gameModes.VIEW) {
                                    if (e.touches.length === 0 && singleTap) {
                                        // check if move can be classified as a
                                        // single tap
                                        if ((touchX - touchStartX) > singleTapSensitivity
                                                || (touchY - touchStartY) > singleTapSensitivity) {
                                            singleTap = false;
                                        }
                                    } else if (e.touches.length === 1) {
                                        var addToDistance = (Math.sqrt(Math.pow(
                                                e.touches[0].clientX - e.changedTouches[0].clientX,
                                                2)
                                                + Math.pow(e.touches[0].clientY
                                                        - e.changedTouches[0].clientY, 2)) - deltaDistance) / 100;
                                        distance += addToDistance;
                                    }
                                }
                                mouseMoveEnd = true;

                            }, false);

            /**
             * Reset body positions and move count
             */
            function resetGame() {

                clearInterval(handler);

                // reset spheres positions and type
                mainSphere[0].SetPosition(new b2Vec2(width / 2, height - mainSphereRadius * 2));
                mainSphere[0].SetLinearVelocity(new b2Vec2(0, 0));

                for ( var i = 0; i < 2; ++i) {
                    spheresP1[i].SetPosition(new b2Vec2(positionP1 + i * playerShift,
                            playerSphereRadius + borderSize));
                    spheresP2[i].SetPosition(new b2Vec2(positionP2 + i * playerShift,
                            playerSphereRadius + borderSize));
                    spheresP1[i].SetUserData(sphereType.PLAYER1_ACTIVE);
                    spheresP2[i].SetUserData(sphereType.PLAYER2_IDLE);
                    spheresP1[i].SetLinearVelocity(new b2Vec2(0, 0));
                    spheresP2[i].SetLinearVelocity(new b2Vec2(0, 0));

                }

                // reset round
                round = 0;
                currentPlayer = player.PLAYER1;
                switchToGameMode();

                // reset 3D WebGL parameters
                distance = 0;
                tiltAngle = 0;
                twistAngle = 0;

            }

            /**
             * Called when touch moves. Save touch coordinates and get chosen
             * body
             */
            function handleTouchMove(e) {
                if (e.touches.length === 1) {
                    e.preventDefault();
                    if (newMouseMove) {
                        body = getBodyAtTouch();
                        if (body) {
                            newMouseMove = false;
                        }
                    }
                    touchX = e.changedTouches[0].pageX / SCALE;
                    touchY = e.changedTouches[0].pageY / SCALE;
                }
                if (gameMode === gameModes.VIEW) {
                    if (e.touches.length === 2) {
                        var newDistance = Math.sqrt(Math.pow(e.touches[0].clientX
                                - e.touches[1].clientX, 2)
                                + Math.pow(e.touches[0].clientY - e.touches[1].clientY, 2));
                        var addToDistance = (newDistance - deltaDistance) / 100;
                        deltaDistance = newDistance;
                        distance += addToDistance;
                    } else if (e.touches.length === 3) {
                        twistAngle -= (e.touches[2].clientX - xPosTwist) / 100;
                        xPosTwist = e.touches[2].clientX;
                        tiltAngle -= (e.touches[2].clientY - yPosTilt) / 100;
                        yPosTilt = e.touches[2].clientY;
                    }
                }
            }

            /**
             * Get b2Body object that was touched
             *
             * @returns Body, that was touched (b2Body object)
             */
            function getBodyAtTouch() {
                var mousePVec = new b2Vec2(touchX, touchY);
                var aabb = new b2AABB();
                aabb.lowerBound.Set(touchX - 1, touchY - 1);
                aabb.upperBound.Set(touchX + 1, touchY + 1);

                // Query the world for overlapping shapes.

                selectedBody = null;
                world.QueryAABB(getBodyCB, aabb);
                return selectedBody;
            }

            /**
             * Get body touched and save it to selectBody variable
             *
             * @param fixture
             *            Touched fixture
             * @returns true if success, false otherwise
             */
            function getBodyCB(fixture) {
                if (fixture.GetBody().GetType() !== b2Body.b2_staticBody
                        && fixture.GetShape().TestPoint(fixture.GetBody().GetTransform(),
                                new b2Vec2(touchX, touchY))) {
                    selectedBody = fixture.GetBody();
                    return false;
                }
                return true;
            }

            /**
             * Check which sphere is being moved. If it can be moved increase
             * move count, else return false
             *
             * @param body
             *            Body to be moved
             * @returns true if body can be moved, false if it can't or if body
             *          is null
             */
            function checkMove(body) {
                if (!body || lastMove || gameMode === gameModes.VIEW) {
                    return false;
                }

                var pos = body.GetPosition();

                // sphere one
                if (pos === spheresP1[0].GetPosition() || pos === spheresP1[1].GetPosition()) {
                    if (currentPlayer === player.PLAYER1) {
                        currentPlayer = player.PLAYER2;
                        spheresP1[0].SetUserData(sphereType.PLAYER1_IDLE);
                        spheresP1[1].SetUserData(sphereType.PLAYER1_IDLE);
                        return true;
                    } else {
                        return false;
                    }
                    // sphere two
                } else if (pos === spheresP2[0].GetPosition() || pos === spheresP2[1].GetPosition()) {
                    if (currentPlayer === player.PLAYER2) {
                        currentPlayer = player.PLAYER1;
                        round++;
                        spheresP2[0].SetUserData(sphereType.PLAYER2_IDLE);
                        spheresP2[1].SetUserData(sphereType.PLAYER2_IDLE);
                        return true;
                    } else {
                        return false;
                    }
                    // main sphere (can't be moved)
                } else {
                    return false;
                }

            }
            /**
             * Check if user used all their moves
             *
             * @returns true if end games, false otherwise
             */
            function checkIsEndGame() {
                if (round >= MAX_MOVES) {
                    return true;
                } else {
                    return false;
                }
            }

            /**
             * Check if any of the bodies is moving
             *
             * @returns True if any of the bodies is moving, false if all the
             *          movies are stopped
             */
            function checkAreBodiesMoving() {

                if (!gameUtils.compareVelocity(spheresP1[0].GetLinearVelocity(), moveTreshold)
                        || !gameUtils.compareVelocity(spheresP1[0].GetLinearVelocity(),
                                moveTreshold)) {
                    return true;
                }
                if (!gameUtils.compareVelocity(spheresP2[0].GetLinearVelocity(), moveTreshold)
                        || !gameUtils.compareVelocity(spheresP2[0].GetLinearVelocity(),
                                moveTreshold)) {
                    return true;
                }
                if (!gameUtils.compareVelocity(mainSphere[0].GetLinearVelocity(), moveTreshold)) {
                    return true;
                }
                return false;
            }

            /**
             * Switch game to play mode. Changes next player to active. Called
             * after single tap if there are still moves to make
             */
            function switchToGameMode() {
                gameMode = gameModes.PLAY;

                if (currentPlayer === player.PLAYER2) {
                    spheresP2[0].SetUserData(sphereType.PLAYER2_ACTIVE);
                    spheresP2[1].SetUserData(sphereType.PLAYER2_ACTIVE);
                } else {
                    if (round < MAX_MOVES) {
                        spheresP1[0].SetUserData(sphereType.PLAYER1_ACTIVE);
                        spheresP1[1].SetUserData(sphereType.PLAYER1_ACTIVE);
                    }
                }

                distance = 0;
                tiltAngle = 0;
                twistAngle = 0;
                $("#bottomBar").text(
                        "Rotate/zoom disabled. Player " + playerColors[currentPlayer] + " move");
            }

            /**
             * Switch game to view mode. Enable zooming and moving the screen.
             */
            function switchToViewMode() {
                gameMode = gameModes.VIEW;
                if (checkIsEndGame()) {
                    $("#bottomBar").text("Rotate/zoom enabled. Wait...");
                } else {
                    $("#bottomBar").text("Rotate/zoom enabled. Tap to continue game");
                }
            }
            /**
             * Update bodies position, game results and show the data to WebGl
             * module
             */
            function update() {
                // when user stopped aiming - move the boule

                if (mouseMoveEnd) {
                    mouseMoveEnd = false;

                    if (checkMove(body)) {
                        lastMove = checkIsEndGame();
                        body.SetLinearVelocity(new b2Vec2(5 * (touchX - touchStartX),
                                5 * (touchY - touchStartY)));
                        body.SetAwake(true);
                        switchToViewMode();
                    }

                    else if (singleTap && !checkIsEndGame()) {
                        switchToGameMode();
                    }

                }

                // if the last move was made - wait for bodies stop and show
                // results
                if (lastMove && !checkAreBodiesMoving()) {
                        var result = gameUtils.checkReults(spheresP1.concat(spheresP2, mainSphere));
                        lastMove = false;
                        resetGame();
                        results.initialize(result);
                }

                // send data to WebGl module
                if (!debug) {

                    var spheres = spheresP1.concat(spheresP2, mainSphere);
                    var i = 0;

                    var sphereData = [];
                    var sphresAngularVelocity = [];
                    for (i = 0; i < spheres.length; i++) {
                        var pos = spheres[i].GetPosition();
                        var velocity = spheres[i].GetLinearVelocity();
                        // newX, newY position values can range between 0 and 1
                        sphereData.push({
                            x : pos.x / width,
                            y : pos.y / height,
                            radius : spheres[i].GetFixtureList().GetShape().GetRadius() / width,
                            type : spheres[i].GetUserData()
                        });
                        sphresAngularVelocity.push({
                            xVelocity : velocity.x,
                            yVelocity : velocity.y,
                            radius : spheres[i].GetFixtureList().GetShape().GetRadius()
                        });
                    }

                    if (sphereData.length) {
                        webgl.drawScene(sphereData, distance, tiltAngle, twistAngle);
                    }
                    webgl.animate(sphresAngularVelocity);
                }

                world.Step(1 / 60, 10, 10);
                world.DrawDebugData();
                world.ClearForces();
            }

        }

    };
}($);