/*******************************************************************************************************************************************************************************************************************************************************************************************************************************************************************************************************************************************************************************************************************
 *
 * @author Tadeusz Wlodarkiewicz <<ahref='mailto:t.wlodarkiew@samsung.com'>t.wlodarkiew@samsung.com</a>>, Karolina Krolik <<ahref='mailto:k.krolik@samsung.com'>k.krolik@samsung.com</a>>
 *
 *
 * ********************************************************************************************************
 *
 * Copyright (c) 2012 Samsung Electronics All Rights Reserved.
 *
 ******************************************************************************************************************************************************************************************************************************************************************************************************************************************************************************************************************************************************************************************************************/

/**
 * The following code was created with help and using source code from Learning WebGL Tutorial Website.
 * Learning WebGL author: Giles Thomas
 * Learning WebGL address: http://learningwebgl.com
 *
 */

/**
 * Handles all webgl functionality, loading buffers, drawing objects, rotation
 * and translation of the objects based on the information received from game
 * module.
 */
var webgl = function($) {
    // reference to webgl
    var gl;
    // reference to shader
    var shaderProgram;

    // model view matrix
    var mvMatrix = mat4.create();

    // model view matrix array
    var mvMatrixStack = [];

    // projection matrix
    var pMatrix = mat4.create();

    // array of spheres rotations
    var rotationSpheres = [];

    // displaying fps
    var date = new Date();

    // FPS value
    var fps = 0;

    // time interval between each FPS display in ms
    var updateInterval = 10000;
    // time since the displayFPS function displayed FPS
    var timeSinceLastUpdate = 0;
    // number of frames that were displayed during timeSinceLastUpdate time
    // interval
    var frameCount = 0;
    // current time
    var currentTick;
    // the last time the displayFPS function was started
    var lastTick;
    // checks if displayFPS function was started for the fist time
    var isFirst = true;
    // the last time the animation was done
    var lastTime = 0;

    /**
     * Changes coordinates from [0..1] to [-1..1]
     *
     * @param x
     *            {Number} x coordinate
     */
    var adjustCoordinateX = function(x) {
        return (2 * x - 1);
    };

    /**
     * Changes coordinates from [0..1] to [-1..1]
     *
     * @param y
     *            {Number} y coordinate
     */
    var adjustCoordinateY = function(y) {
        return (gl.viewportHeight / gl.viewportWidth) * -(2 * y - 1);
    };

    /**
     * Initiates WebGL on a specified canvas and sets the viewport height and
     * width
     */
    var initGL = function(canvas) {
        try {
            gl = canvas.getContext("experimental-webgl");
            gl.viewportWidth = canvas.width;
            gl.viewportHeight = canvas.height;
            console.log("canvas.width: " + canvas.width);
            console.log("canvas.height: " + canvas.height);
        } catch (e) {
            alert("Failed to init WebGL");
        }
        if (!gl) {
            alert("Could not initialise WebGL, sorry :-(");
        }
    };

    /**
     * Gets shader parameters and creates shader
     *
     * @param gl
     *            reference to webgl
     * @param id
     *            of the element where shader is located
     * @returns shader
     */
    var getShader = function(gl, id) {
        var shaderScript = document.getElementById(id);
        if (!shaderScript) {
            return null;
        }

        var str = "";
        var k = shaderScript.firstChild;
        while (k) {
            if (k.nodeType === 3) {
                str += k.textContent;
            }
            k = k.nextSibling;
        }

        var shader;
        if (shaderScript.type === "x-shader/x-fragment") {
            shader = gl.createShader(gl.FRAGMENT_SHADER);
        } else if (shaderScript.type === "x-shader/x-vertex") {
            shader = gl.createShader(gl.VERTEX_SHADER);
        } else {
            return null;
        }

        gl.shaderSource(shader, str);
        gl.compileShader(shader);

        if (!gl.getShaderParameter(shader, gl.COMPILE_STATUS)) {
            alert(gl.getShaderInfoLog(shader));
            return null;
        }

        return shader;
    };

    /**
     * Initiates shaders
     */
    var initShaders = function() {
        var fragmentShader = getShader(gl, "shader-fs");
        var vertexShader = getShader(gl, "shader-vs");

        shaderProgram = gl.createProgram();
        gl.attachShader(shaderProgram, vertexShader);
        gl.attachShader(shaderProgram, fragmentShader);
        gl.linkProgram(shaderProgram);

        if (!gl.getProgramParameter(shaderProgram, gl.LINK_STATUS)) {
            alert("Could not initialise shaders");
        }

        gl.useProgram(shaderProgram);

        shaderProgram.vertexPositionAttribute = gl.getAttribLocation(shaderProgram,
                "aVertexPosition");
        gl.enableVertexAttribArray(shaderProgram.vertexPositionAttribute);

        shaderProgram.textureCoordAttribute = gl.getAttribLocation(shaderProgram, "aTextureCoord");
        gl.enableVertexAttribArray(shaderProgram.textureCoordAttribute);

        shaderProgram.vertexColorAttribute = gl.getAttribLocation(shaderProgram, "aVertexColor");
        gl.enableVertexAttribArray(shaderProgram.vertexColorAttribute);

        shaderProgram.pMatrixUniform = gl.getUniformLocation(shaderProgram, "uPMatrix");
        shaderProgram.mvMatrixUniform = gl.getUniformLocation(shaderProgram, "uMVMatrix");
        shaderProgram.nMatrixUniform = gl.getUniformLocation(shaderProgram, "uNMatrix");
        shaderProgram.samplerUniform = gl.getUniformLocation(shaderProgram, "uSampler");
    };

    var mvPushMatrix = function() {
        var copy = mat4.create();
        mat4.set(mvMatrix, copy);
        mvMatrixStack.push(copy);
    };

    var mvPopMatrix = function() {
        if (mvMatrixStack.length === 0) {
            throw "Invalid popMatrix!";
        }
        mvMatrix = mvMatrixStack.pop();
    };

    var setMatrixUniforms = function() {
        gl.uniformMatrix4fv(shaderProgram.pMatrixUniform, false, pMatrix);
        gl.uniformMatrix4fv(shaderProgram.mvMatrixUniform, false, mvMatrix);

        var normalMatrix = mat3.create();
        mat4.toInverseMat3(mvMatrix, normalMatrix);
        mat3.transpose(normalMatrix);
        gl.uniformMatrix3fv(shaderProgram.nMatrixUniform, false, normalMatrix);
    };

    /**
     * Displays FPS in console
     */
    var displayFPS = function() {

        date = new Date();
        if (isFirst) {
            lastTick = date.getTime();
            isFirst = false;
        }

        frameCount++;
        currentTick = date.getTime();

        var elapsed = currentTick - lastTick;

        lastTick = currentTick;
        timeSinceLastUpdate += elapsed;

        if (timeSinceLastUpdate && timeSinceLastUpdate > updateInterval) {
            fps = (frameCount / timeSinceLastUpdate) * 1000;
            console.log("fps: " + fps);

            frameCount = 0;
            timeSinceLastUpdate -= updateInterval;
        }
    };

    return {

        /**
         * Initiate a sphere with defined radius and type
         *
         * @param radius
         *            {Number} Radius of the sphere
         * @param type
         *            {Number}Type of the sphere
         */
        initSphere : function(radius, type) {
            rotationSpheres[type] = {
                x : 0,
                y : 0
            };
            spheres.createSphere(radius, type);
        },

        /**
         * Draws spheres on the canvas.
         *
         * @param sphereData
         *            {Array} positions of the spheres where the spheres should
         *            be drawn, radius and types of the spheres
         * @param distance
         *            {Number} additional distance of the playing board from the
         *            camera, used to zoom in and out
         * @param tiltAngle
         *            {Number} rotation along the x axis, used to rotate camera
         *            around the board
         * @param twistAngle
         *            {Number} rotation along the z axis, used to rotate camera
         *            around the board
         */
        drawScene : function(sphereData, distance, tiltAngle, twistAngle) {
            // displayFPS();

            gl.viewport(0, 0, gl.viewportWidth, gl.viewportHeight);
            gl.clear(gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT);

            mat4.perspective(45, gl.viewportWidth / gl.viewportHeight, 0.1, 100.0, pMatrix);

            mat4.identity(mvMatrix);
            mat4.translate(mvMatrix, [ 0, 0, -4.0 + distance ]);

            mvPushMatrix();
            ground.loadGround();
            mat4.rotateX(mvMatrix, tiltAngle);
            mat4.rotateZ(mvMatrix, twistAngle);
            mat4.scale(mvMatrix, [ 1, gl.viewportHeight / gl.viewportWidth, 1 ]);
            setMatrixUniforms();
            ground.drawGround();
            mvPopMatrix();

            for ( var i = 0; i < sphereData.length; i++) {
                spheres.loadSphere(sphereData[i].type);

                mvPushMatrix();
                mat4.rotateX(mvMatrix, tiltAngle);
                mat4.rotateZ(mvMatrix, twistAngle);
                mat4.translate(mvMatrix, [ adjustCoordinateX(sphereData[i].x),
                        adjustCoordinateY(sphereData[i].y), 2 * sphereData[i].radius ]);

                mat4.rotateX(mvMatrix, rotationSpheres[i].y);
                mat4.rotateY(mvMatrix, rotationSpheres[i].x);

                mat4.scale(mvMatrix, [ 2, 2, 2 ]);

                setMatrixUniforms();
                spheres.drawSphere(sphereData[i].type);
                mvPopMatrix();
            }
        },

        /**
         * Animating spheres rotation
         *
         * @param sphresAngularVelocity
         *            {Array} array of the angular velocities of the spheres
         *            with their radius. Based on them rotation is calculated
         */
        animate : function(sphresAngularVelocity) {
            var timeNow = new Date().getTime();
            if (lastTime !== 0) {
                var elapsed = timeNow - lastTime;
                for ( var i = 0; i < sphresAngularVelocity.length; i++) {
                    rotationSpheres[i].x += ((sphresAngularVelocity[i].xVelocity / sphresAngularVelocity[i].radius) * elapsed) / 1000.0;
                    rotationSpheres[i].y += ((sphresAngularVelocity[i].yVelocity / sphresAngularVelocity[i].radius) * elapsed) / 1000.0;
                }
            }
            lastTime = timeNow;
        },

        /**
         * Starts WebGL on a specified canvas
         *
         * @param canvas
         */
        webGLStart : function(canvas) {
            console.log("webgl start");
            initGL(canvas);
            initShaders();
            spheres.initSpheres(gl, shaderProgram);
            ground.initGround(gl, shaderProgram);

            gl.clearColor(0.0, 0.0, 0.5, 0.5);
            gl.enable(gl.DEPTH_TEST);
        }

    };

}($);
